#!/usr/bin/env python3

# This library (which can also act as a test client) is written to be able to download specific files from PS
# servers which we need in order to do OSUpgrade and other similar actions
# Right now, the only supported file it downloads is WIM.ini
# This library can be imported by another python script and called to download the WIM.ini to a specified location

import os
import sys
import dns.resolver
import requests
requests.packages.urllib3.disable_warnings(requests.packages.urllib3.exceptions.InsecureRequestWarning)
import socket
import logging
import argparse

# if script's parent directory isn't in sys.path, then add it
if ('/'.join(sys.path[0].split("/")[:-1]) not in sys.path):
    sys.path.append('/'.join(sys.path[0].split("/")[:-1]))

from apsdk import apenv


class PSFileClient(object):
    def __init__(self):
        self.use_certs = False

    # Setup logging
    def log_setup(self) -> bool:
        try:
            logger = logging.getLogger()
            logger.setLevel(logging.INFO)
            formatter = logging.Formatter('%(asctime)s [%(process)d]: %(message)s', '%b %d %H:%M:%S')

            handler_stdout = logging.StreamHandler(sys.stdout)
            handler_stdout.setFormatter(formatter)
            handler_stdout.addFilter(type('', (logging.Filter,), {'filter': staticmethod(lambda r: r.levelno <= logging.INFO)}))
            logger.addHandler(handler_stdout)

            handler_stderr = logging.StreamHandler(sys.stderr)
            handler_stderr.setFormatter(formatter)
            handler_stderr.setLevel(logging.WARNING)
            logger.addHandler(handler_stderr)
        except Exception as e:
            print("psfileclient.log_setup: Failed to setup logging. Error: {0}".format(e), file=sys.stderr)
            return False

        return True

    # Cofigure whether to use MF certs for authentication or not
    def set_use_certs(self, use_certs: bool=False):
        if use_certs:
            logging.debug("Using https certs: True")
            self.use_certs = True
        else:
            self.use_certs = False
            logging.debug("Using https certs: False")

    # Get current hostname
    def get_hostname(self) -> tuple:
        try:
            hostname = socket.gethostname()
        except Exception as e:
            logging.error("psfileclient.get_hostname: Failed to get hostname via socket.gethostname(). Error: {0}".format(e))
            return "", False

        logging.debug("Hostname: {0}".format(hostname))
        return hostname, True

    # Get the cluster name
    def get_cluster(self) -> tuple:
        try:
            cluster = apenv.get_cluster_name().lower()
        except Exception as e:
            logging.error("psfileclient.get_cluster: Failed to get cluster name. Error: {0}".format(e))
            return "", False

        logging.debug("Cluster: {0}".format(cluster))
        return cluster, True

    # Get the current machine environment
    def get_environment(self) -> tuple:
        try:
            environment = apenv.get_environment_name().lower()
        except Exception as e:
            logging.error("psfileclient.get_environment: Failed to get environment name. Error: {0}".format(e))
            return "", False

        logging.debug("Environment: {0}".format(environment))
        return environment, True

    # Get list of ps servers
    def get_ps_servers(self, cluster: str) -> tuple:
        psvip = "all.ps.autopilot.{0}.ap.gbl".format(cluster)
        logging.debug("psvip: {0}".format(psvip))

        try:
            servers = dns.resolver.query(psvip, 'A')
        except Exception as e:
            logging.error("psfileclient.get_ps_servers: Failed to get PS servers. Error: {0}".format(e))
            return "", False

        psservers = []
        if len(servers) > 0:
            for server in servers:
                psservers.append(str(server))
        else:
            logging.error("psfileclient.get_ps_servers: No PS servers found in the returned DNS query.")
            return "", False

        return psservers, True

    # Get the machine function cert/key pair as a tuple
    def get_machine_function_cert(self) -> tuple:
        try:
            machine_function_cert = apenv.get_machine_function_cert()
            if machine_function_cert is None:
                return "", ""
        except Exception as e:
            logging.error("psfileclient.get_machine_function_cert: Could not obtain machine function cert paths.")
            logging.error("psfileclient.get_machine_function_cert: Error: {0}".format(e))
            return "", ""

        return machine_function_cert

    # Build PS server filepath
    def get_filepath(self, psserver: str, filename: str):
        if self.use_certs:
            http_option = "https://"
            port = 6698
        else:
            http_option = "http://"
            port = 6798

        logging.debug("HTTP_Option: {0}".format(http_option))
        logging.debug("Port: {0}".format(port))
        logging.debug("PSServer: {0}".format(psserver))

        filepath = ""

        if filename == "wim.ini":
            filepath = "{0}{1}:{2}/data/provisioning/data/images/wim/wim.ini".format(http_option, psserver, port)

        logging.debug("Filepath: {0}".format(filepath))

        return filepath, True

    # Download the PS file
    def get_ps_file(self, psservers: list, filename: str) -> tuple:
        machine_function_cert = None
        if self.use_certs:
            machine_function_cert = self.get_machine_function_cert()
            if not machine_function_cert[0] or not machine_function_cert[1]:
                logging.error("psfileclient.get_ps_file: One or both of the machine function cert paths is missing.")
                return "", False

        response = None

        for psserver in psservers:
            filepath, success = self.get_filepath(psserver, filename)
            if not success:
                return "", False

            try:
                response = requests.get(filepath,
                                        cert=machine_function_cert,
                                        verify=False)
            except requests.exceptions.ConnectTimeout:
                logging.error("psfileclient.get_ps_file: ConnectTimeout: {0}".format(filepath))
                return "", False

            except requests.exceptions.ConnectionError:
                logging.error("psfileclient.get_ps_file: ConnectionError: {0}".format(filepath))
                return "", False

            except requests.exceptions.Timeout:
                logging.error("psfileclient.get_ps_file: Timeout: {0}".format(filepath))
                return "", False

            except Exception as e:
                logging.error("psfileclient.get_ps_file: Could not download PS file")
                logging.error("psfileclient.get_ps_file: Error: {0}".format(e))
                return "", False

            if not response.content:
                continue
            else:
                break

        if isinstance(response, type(None)):
            logging.error("psfileclient.get_ps_file: No response from PS servers")
            return "", False

        if isinstance(response.content, bytes):
            logging.debug(response.content)
            return response.content, True
        else:
            logging.error("psfileclient.get_ps_file: Response format from PS server was not of type bytes")
            return "", False

    # Decode binary data from http to string
    def decode_file_data(self, content: bytes, encoding: str="utf-8") -> tuple:
        decoded_content = None

        try:
            decoded_content = content.decode(encoding)
        except Exception as e:
            logging.error("psfileclient.decode_file_data: Could not decode binary file data to string using: {0}".format(encoding))
            logging.error("psfileclient.decode_file_data: Error: {0}".format(e))
            return "", False

        if not isinstance(decoded_content, str):
            logging.error("psfileclient.decode_file_data: Binary data did not successfully decode to type(str)")
            return "", False

        file_content = decoded_content.split("\r\n")

        return file_content, True

    # Write the ps file to the output destination on disk
    def write_ps_file(self, file_content: list, output_filename: str) -> bool:
        output_directory = os.path.dirname(output_filename)
        if not os.path.isdir(output_directory):
            logging.error("psfileclient.write_ps_file: The output filename provided must be in a directory that already exists. Please create the directory first")
            return False

        logging.debug("Writing file: {0}".format(output_filename))

        try:
            with open(output_filename, 'w') as outfile:
                for line in file_content:
                    outfile.write("{0}\n".format(line))
        except EnvironmentError as e:
            logging.error("psfileclient.write_ps_file: Returned error: {0}".format(e.strerror))
            return False

        return True

    # Function which does the main work which can be imported and used by other files as a library so that it can return
    # a bool result for other python files using it
    # We return the result to main() which then exits with the appropriate error code
    def process_psserver_file_action(self, filename: str, output_filename: str, use_logging: bool=True, use_certs: bool=False) -> bool:
        # Setup logging
        if use_logging:
            if not self.log_setup():
                print("psfileclient.main: Failed to get setup logging (main.log_setup)", file=sys.stderr)
                return False

        logging.info("psfileclient.main: Starting psfileclient to download file {0} to {1}".format(filename, output_filename))

        # Cofigure whether to use MF certs for authentication or not
        self.set_use_certs(use_certs)

        # Get current hostname
        hostname, success = self.get_hostname()
        if not success:
            logging.error("psfileclient.main: Failed to get hostname (main.get_hostname)")
            return False

        # Get the cluster name
        cluster, success = self.get_cluster()
        if not success:
            logging.error("psfileclient.main: Failed to get cluster (main.get_cluster)")
            return False

        # Get the current machine environment
        environment, success = self.get_environment()
        if not success:
            logging.error("psfileclient.main: Failed to get environment (main.get_environment)")
            return False

        # Get list of ps servers
        psservers, success = self.get_ps_servers(cluster)
        if not success:
            logging.error("psfileclient.main: Failed to get hostname (main.get_ps_servers)")
            return False

        # Download the PS file
        content, success = self.get_ps_file(psservers, filename)
        if not success:
            logging.error("psfileclient.main: Failed to get psserver response (main.get_ps_file)")
            return False

        # Decode binary data from http to string
        file_content, success = self.decode_file_data(content)
        if not success:
            logging.error("psfileclient.main: Failed to decode file data from binary to string (main.decode_file_data)")
            return False

        # Write the ps file to the output destination on disk
        if not self.write_ps_file(file_content, output_filename):
            logging.error("psfileclient.main: Failed to write file data to {0} (main.write_ps_file)".format(output_filename))
            return False

        logging.info("psfileclient.main: Completed psfileclient to download file {0} to {1}".format(filename, output_filename))
        return True

    # Perform the main action of the client and exit appropriately based on result
    def main(self, filename: str, output_filename: str, use_logging: bool=True, use_certs: bool=False):
        if not self.process_psserver_file_action(filename, output_filename, use_logging, use_certs):
            sys.exit(1)
        else:
            sys.exit(0)


# This client can be called as a standalone python script by other programs or imported as a library and used
# accordingly by other python scripts
if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="Get a file from PS servers")
    parser.add_argument("-f", "--filename", metavar="FILENAME", help="File to upload to the server (cannot be empty)")
    parser.add_argument("-o", "--output_filename", metavar="OUTPUT_FILENAME", help="Filename to output to if different than current folder")

    args = parser.parse_args()

    filename = None
    if args.filename:
        filename = args.filename.strip()
        if filename != "wim.ini":
            parser.error("Only wim.ini is currently supported")
    else:
        parser.error("You must specify a filename to fetch from PS servers")

    output_filename = None
    if args.output_filename:
        output_filename = args.output_filename.strip()
        output_directory = os.path.dirname(output_filename)
        if not os.path.isdir(output_directory):
            parser.error("The output filename provided must be in a directory that already exists. Please create the directory first")
    else:
        parser.error("You must specify an output filename to save to")

    PSFileClient().main(filename, output_filename)
