#!/usr/bin/env python3

# This script does NOT initiate any repairs of its own!  It only confirms completion of those repairs to DM.
# This script checks to see if the customer or another repair service has issued a HardReboot repair for the
#   AP Linux machine.
# If it finds a repair waiting to be acknowledged, it notifies DM that the repair has been completed.
# In this case since we only support HardReboot, it will confirm to DM that the machine did indeed reboot successfully
#   AFTER the power was reset on the machine.
# If DM does not get this confirmation, it continues to additional, more invasive repairs such as re-imaging

import os
import sys
import csv
import socket
import requests
import datetime
import subprocess

# if script's parent directory isn't in sys.path, then add it
if ('/'.join(sys.path[0].split("/")[:-1]) not in sys.path):
    sys.path.append('/'.join(sys.path[0].split("/")[:-1]))

from apsdk import apenv
from apsdk import dmclient

# The default umask after AzSecPack remediation is 077 which is too restrictive in cases where we want
# services to make their files visible publicly by default without having to set permissions on them after
# we create them.  Setting the umask to 022 at the beginning of a script makes sure the default behavior is
# default Linux permissions.  If permissions need to be specific, we can set them using python functions.
os.umask(0o022)


class WatchAPRepair(object):
    DEBUG = False
    EAP = False
    EAP_PATH = "/opt/ap/EAP"
    REPAIR_TYPES = ["HardReboot"]
    REPAIR_AFTER_REBOOT = ["HardReboot"]

    # Determine if the machine is EAP or not
    # EAP does its own repairs so we don't need it on EAP
    def is_eap_machine(self) -> bool:
        if os.path.exists(self.EAP_PATH):
            return True

        return False

    # Get the Machine Function
    def get_machine_function(self) -> str:
        machinefunction = apenv.get_machine_function()

        return machinefunction

    # Get hostname
    def get_hostname(self) -> str:
        hostname = socket.gethostname()

        return hostname

    # Get DM Gateway URL
    def get_dm_gateway_url(self) -> str:
        dm_gateway_url = apenv.get_secure_dm_gateway_url()

        return dm_gateway_url

    # Get dm_client object
    def create_dm_client(self, dm_gateway_url) -> dmclient:
        dm_client = dmclient.DMClient(dm_gateway_url)

        return dm_client

    # Get the repair list from DM
    def get_repairs_list_from_dm(self, machine_function, dm_client) -> requests:
        command = "GetRepairList -fields RepairActionState,MachineName,RequestType,RequestTime -f " + machine_function
        response = dm_client.send_request(command, '')

        if self.DEBUG:
            print("WatchAPRepair:get_repairs_list_from_dm DM response: {0}".format(response))
            print("WatchAPRepair:get_repairs_list_from_dm DM response text: \n{0}".format(response.text))

        return response

    # Verify that we got a valid DM response
    def check_dm_response(self, repairs_list: str) -> bool:
        if repairs_list.lower().startswith("no response"):
            print("WatchAPRepair:check_dm_response Got 'No Response' response from DM", file=sys.stderr)
            return False

        return True

    # Check if a DM Repair is required for this host
    def check_if_repair_completion_required(self, repairs_list: str, hostname: str) -> tuple:
        repair = ""
        repair_request_time_utc_str = ""
        lines = repairs_list.splitlines()
        reader = csv.reader(lines)
        for row in reader:
            if len(row) == 4 and row[1].strip() == hostname:
                repair = row[2].strip()
                repair_request_time_utc_str = row[3].strip()
                break

        if not repair:
            return "", ""

        print("WatchAPRepair:check_if_repair_completion_required Repair found: Repair={0}".format(repair))

        # Does the repair type match an existing repair
        if repair in self.REPAIR_TYPES:
            print("WatchAPRepair:check_if_repair_completion_required Will attempt to complete: Repair={0}".format(repair))

        return repair, repair_request_time_utc_str

    # Get the last boot UTC time
    def get_boot_time_utc(self) -> tuple:
        bootup_time_utc = None
        result = None

        local_now = datetime.datetime.now()
        utc_now = datetime.datetime.utcnow()
        utc_offset = utc_now - local_now

        # 2018-09-15 00:10:07
        command = "uptime -s"

        try:
            if self.DEBUG:
                print("WatchAPRepair:get_boot_time_utc Command: {0}".format(command))
            result = subprocess.check_output(command, timeout=60, universal_newlines=True, shell=True)
        except subprocess.CalledProcessError as e:
            print("Returned error: {0}".format(e.stderr), file=sys.stderr)
            return "", False

        try:
            bootup_time_utc = datetime.datetime.strptime(result.strip(), "%Y-%m-%d %H:%M:%S") + utc_offset
            if self.DEBUG:
                print("WatchAPRepair:get_boot_time_utc boot UTC datetime is: {0}".format(bootup_time_utc))
        except Exception as e:
            print("WatchAPRepair:get_boot_time_utc Failed to convert boot time to UTC: {0}".format(e), file=sys.stderr)
            return "", False

        return bootup_time_utc, True

    # Get the repair UTC time
    def get_repair_request_time_utc(self, repair_request_time_utc_str: str) -> tuple:
        try:
            repair_request_time_utc = datetime.datetime.strptime(repair_request_time_utc_str, "%Y-%m-%d %H:%M:%S.%f")
            if self.DEBUG:
                print("WatchAPRepair:get_boot_time_utc Repair request UTC datetime is: {0}".format(repair_request_time_utc))
        except Exception as e:
            print("WatchAPRepair:get_boot_time_utc Failed to convert repair request time to UTC: {0}".format(e), file=sys.stderr)
            return "", False

        return repair_request_time_utc, True

    # Make sure that repairs are completed only after they should have run
    def verify_completion_utc_time(self, repair, boot_time_utc: datetime.datetime, repair_request_time_utc: datetime.datetime) -> bool:
        if repair in self.REPAIR_AFTER_REBOOT:
            if repair_request_time_utc < boot_time_utc:
                return True
            else:
                if self.DEBUG:
                    print("WatchAPRepair:verify_completion_utc_time Repair UTC time is greater or equal to Boot UTC time")

                # Indicate that there's a pending repair which can't be completed yet
                print("WatchAPRepair:verify_completion_utc_time The following is pending but cannot be completed yet: Repair={0}".format(repair))

        return False

    # Build the repair request if repair is necessary
    def build_repair_request(self, repair: str, hostname: str) -> str:
        update_repair_status_upload_data = "#Fields:MachineName,RequestType,ResultCode,MachineEnv,RequestTime,RequestTry,RequestSource,RequestReason,ResultTime,ResultWdCode,ResultSource,ResultDescription,ActionTime,ActionCode,ActionTry,ActionSource,ActionDescription\r\n"
        update_repair_status_upload_data = update_repair_status_upload_data + "{0},{1},Ok,,,,,,,,,,,,,,".format(hostname, repair)

        if self.DEBUG:
            print("WatchAPRepair:build_repair_request DM repair request: \n{0}".format(update_repair_status_upload_data))

        return update_repair_status_upload_data

    # Send the prepared completion request to DM via dmclient
    def send_repair_completion_request_to_dm(self, dm_client: dmclient, update_repair_status_upload_data):
        response = dm_client.send_request("UpdateRepairStatus", update_repair_status_upload_data)

        if self.DEBUG:
            print("WatchAPRepair:send_repair_completion_request_to_dm DM response: {0}".format(response))
            print("WatchAPRepair:send_repair_completion_request_to_dm DM response text: \n{0}".format(response.text))

        return response

    # Verify repair completion response
    def verify_repair_completion_response(self, response: str) -> bool:
        if not response.startswith("OK"):
            return False
        else:
            return True

    def main(self):
        # Determine if the machine is EAP or not
        # EAP does its own repairs so we don't need it on EAP
        if self.is_eap_machine():
            print("watchaprepair.main: Repair service not used on EAP machines (main.is_eap_machine)")
            sys.exit(0)

        # Get machine function
        machine_function = self.get_machine_function()
        if not machine_function:
            print("watchaprepair.main: Could not get machine function (main.get_machine_function)", file=sys.stderr)
            sys.exit(1)

        # Get hostname
        hostname = self.get_hostname()
        if not hostname:
            print("watchaprepair.main: Could not get hostname (main.get_hostname)", file=sys.stderr)
            sys.exit(1)

        # Get dmgateway_url
        dm_gateway_url = self.get_dm_gateway_url()
        if not dm_gateway_url:
            print("watchaprepair.main: Could not get dm_gateway_url (main.get_dm_gateway_url)", file=sys.stderr)
            sys.exit(1)

        # Create dmclient object
        dm_client = self.create_dm_client(dm_gateway_url)
        if not dm_client:
            print("watchaprepair.main: Could not get dm_client (main.create_dm_client)", file=sys.stderr)
            sys.exit(1)

        # Get repairs list from DM
        repairs_list = self.get_repairs_list_from_dm(machine_function, dm_client)
        if not repairs_list:
            print("watchaprepair.main: Could not get repairs_list (main.get_repairs_list_from_dm)", file=sys.stderr)
            sys.exit(1)

        # Verify that the DM response is good
        if not self.check_dm_response(repairs_list.text):
            print("watchaprepair.main: Did not get a valid DM response (main.check_dm_response)", file=sys.stderr)
            sys.exit(1)

        # Check if a DM Repair is required for this host
        repair, repair_utc_time_str = self.check_if_repair_completion_required(repairs_list.text, hostname)
        if not repair:
            print("WatchAPRepair:check_if_repair_completion_required No repair needed")
            sys.exit(0)

        # Get the last boot UTC time
        boot_time_utc, success = self.get_boot_time_utc()
        if not success:
            print("watchaprepair.main: Failed to get last boot time utc (main.get_boot_time_utc)", file=sys.stderr)
            sys.exit(1)

        # Get the repair request UTC time
        repair_request_time_utc, success = self.get_repair_request_time_utc(repair_utc_time_str)
        if not success:
            print("watchaprepair.main: Failed to get repair request UTC time (main.get_repair_request_time_utc)", file=sys.stderr)
            sys.exit(1)

        # Make sure that repairs are completed only after they should have run
        if not self.verify_completion_utc_time(repair, boot_time_utc, repair_request_time_utc):
            print("WatchAPRepair:verify_completion_utc_time No repair needed at this time")
            sys.exit(0)

        # Build the repair request if repair is necessary
        update_repair_status_upload_data = self.build_repair_request(repair, hostname)
        if not update_repair_status_upload_data:
            print("watchaprepair.main: Failed to build a repair request (main.build_repair_request)", file=sys.stderr)
            sys.exit(1)

        # Send the prepared completion request to DM via dmclient
        response = self.send_repair_completion_request_to_dm(dm_client, update_repair_status_upload_data)
        if not response:
            print("WatchAPRepair:main UpdateRepairStatus command invocation failed (main.send_repair_completion_request_to_dm)", file=sys.stderr)
            print("WatchAPRepair:main DM response: {0} (main.send_repair_completion_request_to_dm)".format(response), file=sys.stderr)
            print("WatchAPRepair:main DM response text: {0} (main.send_repair_completion_request_to_dm)".format(response.text), file=sys.stderr)
            sys.exit(1)

        if not self.verify_repair_completion_response(response.text):
            print("WatchAPRepair:main UpdateRepairStatus command invocation failed (main.verify_repair_completion_response)", file=sys.stderr)
            print("WatchAPRepair:main DM response: {0} (main.verify_repair_completion_response)".format(response), file=sys.stderr)
            print("WatchAPRepair:main DM response: {0} (main.verify_repair_completion_response)".format(response.text), file=sys.stderr)
            sys.exit(1)
        else:
            print("WatchAPRepair:complete_repair Succeeded to update repair status of {0} on {1}".format(repair, hostname))
            sys.exit(0)


if __name__ == '__main__':
    WatchAPRepair().main()
